# -*- coding: utf-8 -*-
"""
Created on Tue Aug 20 09:54:12 2019

@author: Thorkild Guldager Sørensen, tgs@emd.dk
         Tobias Ahsbahs, ta@emd.dk

DESCRIPTION:
   Example script to setup a windPRO project purely from online data.
   Terrain and roughness data are downloaded.
   An observed wind climate is added from ERA5 data.
   A WAsP calculation is prepared.
   Four turbines are added and a PARK calculation is perfomed.
   Results are exported to csv and pdf.

USAGE:
   Get a working example how to use windPRO scripting and try it out.

NOTES:
   This is for version windPRO version 3.5
   You might need to change some path to windPRO and make a new folder for the project.

REFERENCES
"""

import os
import windproapi.windproapi
from windproapi.utils import get_windpro_sample_path
from windproapi import WindProApi
from windproapi import nan_to_skipvalue

_windproapi = WindProApi()
_windproapi.start_windpro_random_port()

# Paths for this project
working_dir = os.path.join(get_windpro_sample_path('3.6'), 'Scripting Generated Project')
os.makedirs(working_dir, exist_ok=True)
project_path = os.path.join(working_dir, 'Scripting Generated Project.w36p')

# Get to all services
online_data_service = _windproapi.get_service('OnlineDataService')
project_service = _windproapi.get_service('ProjectService')
objects_service = _windproapi.get_service('ObjectsService')
calculation_service = _windproapi.get_service('CalculationService')
calc_statgen_service = _windproapi.get_service('CalcStatgenService')
calc_park_service = _windproapi.get_service('CalcParkService')
calculation_service = _windproapi.get_service('CalculationService')
obj_site_data_service = _windproapi.get_service('ObjSiteDataService')
obj_meteo_service = _windproapi.get_service('ObjMeteoService')
obj_wtg_service = _windproapi.get_service('ObjWtgService')
factory = _windproapi.get_factory('WindproService')

# Position of the project
# This position can be anywhere in the world

# Denmark example
# lng = 9.382716
# lat = 56.627608

# US example
lng = -101.6
lat = 46.791551

# Make a new project
project_service.NewProject(lng=lng, lat=lat, filename=project_path)

# Setting the project coordinate system to UTM
utm_epsg = project_service.GetUtmWgs84EpsgForProject()
project_service.SetProjectCoorEPSG(epsg=utm_epsg)

# Map data #
# Adding a layer for terrain data
terrain_layer = objects_service.AddLayer(layerName='Terrain data', parentFolderHandle=0)
objects_service.SetEditLayer(handle=terrain_layer)

# Getting elevation data sets
terrain_data_type = 'ODSSurfaceGrid'
online_data_service.PrepareService(dataType=terrain_data_type, lat=lat, lon=lng)
terrain_services = online_data_service.GetServices(dataType=terrain_data_type)
print(terrain_services)
# Choosing a specific dataset and downloading the data
terrain_service_id = 'Global_TandemX90_Grid'
terrain_handle = online_data_service.DownloadHeightData(dataType=terrain_data_type,
                                                        implId=terrain_service_id,
                                                        userDesc='Terrain data grid',
                                                        lat=lat,
                                                        lon=lng,
                                                        width=20000,
                                                        height=20000,
                                                        useAsTin=True)

# Getting roughness dataset sets
roughness_data_type = 'ODSRoughnessGridToLine'
online_data_service.PrepareService(dataType=roughness_data_type, lat=lat, lon=lng)
roughness_service = online_data_service.GetServices(dataType=roughness_data_type)
print(roughness_service)
# Choosing a specific roughness dataset and downloading it
roughness_service_id = 'DataService_Rou_grid_GlobCover'
roughness_handle = online_data_service.DownloadRoughnessData(dataType=roughness_data_type,
                                                             implId=roughness_service_id,
                                                             userDesc='Roughness data',
                                                             lat=lat,
                                                             lon=lng,
                                                             width=20000,
                                                             height=20000)

# Meteorological data #
# making a new layer for meteodata
meteo_layer = objects_service.AddLayer(layerName='Reference meteo data', parentFolderHandle=0)
objects_service.SetEditLayer(handle=terrain_layer)

# Getting climate data sets
meteo_data_type = 'ODSClimate'
online_data_service.PrepareService(dataType=meteo_data_type, lat=lat, lon=lng)
meteo_services = online_data_service.GetServices(dataType=meteo_data_type)
print(meteo_services)
# Selecting a specific dataset and downloading it
meteo_service_id = 'siEra5Basic'
meteo_handles = online_data_service.DownloadMeteoData(implId=meteo_service_id,
                                                      lat=lat,
                                                      lon=lng,
                                                      maxDist=30000,
                                                      numPoints=1,
                                                      fromYear=2014,
                                                      toYear=2015)
# Using the Meteo Service to get the
meteo_obj_era5 = obj_meteo_service.GetMeteoObject(handle=meteo_handles[0])

# Flow modelling #
# Addign a layer of
site_data_layer = objects_service.AddLayer(layerName='Site data objects', parentFolderHandle=0)
objects_service.SetEditLayer(handle=site_data_layer)

# Adding a site data object
obj_site_data_statgen = objects_service.AddObject(apiObjType='SiteData',
                                                  lat=lat,
                                                  lng=lng,
                                                  userDesc='STATGEN SiteData')
obj_site_data_service.SetPurpose(handle=obj_site_data_statgen.Handle, purpose='SdPurpStatgen')
# Connecting elevation and roughness data and
obj_site_data_service.TerrainLinkElevationAndRoughnessLine(handle=obj_site_data_statgen.Handle,
                                                           elevHandle=terrain_handle,
                                                           rouLineHandle=roughness_handle)
# Getting an update of the changes that have been made to the object back to python
obj_site_data_statgen = obj_site_data_service.GetSiteDataObject(obj_site_data_statgen.Handle)
obj_site_data_service.ExportCombinedMapFile(obj_site_data_statgen.Handle, 0., 0., 0., 0.,
                                            os.path.join(working_dir, 'testtgs.map'))

# Creating a new statgen calculation calculation and starting to edit it
wws_file = os.path.join(working_dir, 'wind_stat.wws')
statgen_handle = calculation_service.CreateEmpty(calcType='CalcStatgen')
calculation_service.OpenCalcForEdit(handle=statgen_handle)
# Getting the statgen calculation object and adding information on
statgen = calc_statgen_service.GetStatgenCalc()
statgen.Filename = wws_file
statgen.Name = 'Statgen for test'
statgen.SiteDataHandle = obj_site_data_statgen.Handle
statgen.MeteoHandle = meteo_obj_era5.Handle
# Within the meteo object each height has a handle to choose the measurement signal
statgen.MeteoHeiHandle = 2
nan_to_skipvalue(statgen)
calc_statgen_service.SetStatgenCalc(statgen)
calculation_service.CloseCalc(save=True)
calculation_service.Calculate(handle=statgen_handle)

# Saving all the results in the statgen calcualtion
statgenResToFiles = calculation_service.GetResToFiles(handle=statgen_handle)
calculation_service.ResultToFile(handle=statgen_handle,
                                 id=statgenResToFiles[0]['ID'],
                                 filename=os.path.join(working_dir, 'statgen_restofile.xml'))

# Getting reports from statgen. All reports are saved individually.
statgenReports = calculation_service.GetReports(handle=statgen_handle)
for rep in statgenReports:
    calculation_service.SaveReportAsPdf(handle=statgen_handle, id=rep['ID'],
                                        filename=os.path.join(working_dir, 'statgen_' + rep['Name'] + '.pdf'))

# WAsP calculation #
siteDataWasp = objects_service.AddObject(apiObjType='SiteData',
                                         lat=lat,
                                         lng=lng,
                                         userDesc='WASP SiteData')
obj_site_data_service.SetPurpose(handle=siteDataWasp.Handle, purpose='SdPurpWasp')
obj_site_data_service.TerrainLinkElevationAndRoughnessLine(handle=siteDataWasp.Handle,
                                                           elevHandle=terrain_handle,
                                                           rouLineHandle=roughness_handle)
obj_site_data_service.SetWindstatFn(handle=siteDataWasp.Handle,
                                    fileName=wws_file)

# Wind Turbines #
# Path with wind turbine data
wtgPath = os.path.join(os.getcwd(), 'SIEMENS SWT-2.3_test.wtg')

# Loading data, naming it and giving it a location
wtg_lats = [lat - 0.005, lat + 0.01, lat + 0.005, lat - 0.01]
wtg_lngs = [lng - 0.005, lng - 0.01, lng + 0.005, lng + 0.01]

for i, i_lat, i_lng in zip(range(4), wtg_lats, wtg_lngs):
    new_obj = objects_service.AddObject(apiObjType='NewWTG',
                                        lat=i_lat,
                                        lng=i_lng,
                                        userDesc='New WTG obj')
    wtgObj = obj_wtg_service.GetWtgObject(new_obj.Handle)
    wtgObj.Filename = wtgPath
    wtgObj.Hubheight = 92.6
    wtgObj.UserDescription = f'Turbine_{i}'
    wtgObj.UserLabel = 'Turbine ' + str(i + 1)
    nan_to_skipvalue(wtgObj)
    res = obj_wtg_service.SetWtgObject(wtgObj)


# Creating a new PARK calculation and opening it for editing
parkHandle = calculation_service.CreateEmpty(calcType='CalcPark')
calculation_service.OpenCalcForEdit(handle=parkHandle)
park_calc = calc_park_service.GetParkCalc()
park_calc.ParkCalcType = 'ParkTypeSiteData'
park_calc.Name = 'Park for test'

# Factory is needed to communicate more complex structures like arrays to the API.
# Adding the statgen object for connecting to WAsP
intArr = factory.TApiObjectHandles()
intArr['int'].append(siteDataWasp.Handle)
park_calc.StatData.SiteDataObjects = intArr

# Making an array of WTGs that can be added to the park
wtgids = factory.TApiWtgIds()
for w in objects_service.GetObjects(apiObjType='NewWTG'):
    dummy = calc_park_service.GetTApiWtgId()
    dummy.Handle = w.Handle
    dummy.Rowindex = 0
    wtgids['TApiWtgId'].append(dummy)

# Adding turbines
park_calc.NewWtgs = wtgids
park_calc.ExistWtgs = factory.TApiWtgIds()

# This is for setting all the NaNs in the data to a value that can be received by the API
nan_to_skipvalue(park_calc)

# Sending the PARK calculation back to windPRO
calc_park_service.SetParkCalc(park_calc)
calculation_service.CloseCalc(save=True)

# Calculating PARK calculation
calculation_service.Calculate(handle=parkHandle)

# Exporting results and reports
parkResToFiles = calculation_service.GetResToFiles(handle=parkHandle)
for res in parkResToFiles:
    calculation_service.ResultToFile(handle=parkHandle,
                                     id=res['ID'],
                                     filename=os.path.join(working_dir, 'park_' + res['ID'] + '.txt'))

parkReports = calculation_service.GetReports(handle=parkHandle)
# All reports are exported as single pdfs
for rep in parkReports:
    calculation_service.SaveReportAsPdf(handle=parkHandle, id=rep['ID'],
                                        filename=os.path.join(working_dir, 'park_' + rep['Name'] + '.pdf'))
# All reports are exported in a combined pdf
calculation_service.SaveReportAsPdf(handle=parkHandle,
                                    id=-1,
                                    filename=os.path.join(working_dir, 'park_full_report.pdf'))

# Saving project file
project_service.SaveToFile(filename=project_path)
