import time
import os
import numpy as np
from windproapi import WindProApi
from windproapi.utils import get_windpro_sample_path
from windproapi import nan_to_skipvalue


def find_closest_height_handle(meteo_object, measurement_height):
    # Determining which height is closest to the selection height
    meteo_height_diffs = []
    meteo_heights = []
    meteo_height_handles = []
    for temp in meteo_object.Heights.TApiMeteoHeight:
        meteo_height_diffs.append(temp.Height - measurement_height)
        meteo_heights.append(temp.Height)
        meteo_height_handles.append(temp.Handle)

    index_min = np.argmin(abs(np.array(meteo_height_diffs)))
    meteo_height_handle = meteo_height_handles[index_min]
    return meteo_height_handle


# Opening windPRO
_windproapi = WindProApi()

# Project file
working_dir = os.path.join(get_windpro_sample_path('3.6'), 'New Salem\\3.6')
project_path = os.path.join(working_dir, 'New Salem.w36p')

_windproapi.start_windpro_random_port()

# Services
project_service = _windproapi.get_service('ProjectService')
objects_service = _windproapi.get_service('ObjectsService')
calculation_service = _windproapi.get_service('CalculationService')
calc_park_service = _windproapi.get_service('CalcParkService')
obj_wtg_service = _windproapi.get_service('ObjWtgService')
calc_MCP_service = _windproapi.get_service('CalcMCPService')
obj_meteo_service = _windproapi.get_service('ObjMeteoService')
calc_statgen_service = _windproapi.get_service('CalcStatgenService')
obj_site_data_service = _windproapi.get_service('ObjSiteDataService')
factory = _windproapi.get_factory('WindproService')

# Open project file
project_service.LoadFromFile(filename=project_path)

# Making a new MCP calculation
resHandle = calculation_service.CreateEmpty(calcType='CalcMCP')
calculation_service.OpenCalcForEdit(handle=resHandle)
mcp_calc = calc_MCP_service.GetMCPCalc()

# There is not much in the calculation yet without making a new MCP session
print(mcp_calc)
calc_MCP_service.AddSession()

# With the new session much more information in .Session
mcp_calc = calc_MCP_service.GetMCPCalc()
mcp_calc.Name = 'Generate by scripting'
print(mcp_calc)

# Adding meteorological data
meteo_objs = objects_service.GetObjects(apiObjType='MeteoObjectData')
obj_measurement = [d for d in meteo_objs if d.UserDescription == 'New Salem South'][0]
obj_reference = [d for d in meteo_objs if 'EmdWrf' in d.UserDescription][0]

meteo_measurement = obj_meteo_service.GetMeteoObject(handle=obj_measurement.Handle)
meteo_reference = obj_meteo_service.GetMeteoObject(handle=obj_reference.Handle)

# Getting the handles for the correct heights inside the meteo object
internal_measurement_handle = find_closest_height_handle(meteo_measurement, 60)
internal_reference_handle = find_closest_height_handle(meteo_reference, 60)

ses = mcp_calc.Sessions.TApiMcpSession[0]
ses.Enabled = True
ses.MeasureMeteoHandle = meteo_measurement.Handle
ses.MeasureMeteoUID = internal_measurement_handle
ses.ReferenceMeteoHandle = meteo_reference.Handle
ses.ReferenceMeteoUID = internal_reference_handle

# Site data object for making wind statistics
site_objs = objects_service.GetObjects(apiObjType='SiteData')
sitedata = [d for d in site_objs if d.UserDescription == 'for STATGEN'][0]

# Check that the site data object has the right purpose. As this is for generating wind statistics it needs to be
# SdPurpStatgen
if obj_site_data_service.GetSiteDataObject(sitedata.Handle).Purpose == 'SdPurpStatgen':
    print('Correct purpose for added site data object.')
else:
    raise ValueError('Incorrect purpose for site data objects.')
ses.SGSiteDataHandle = sitedata.Handle

# Adding all MCP models
template_models = calc_MCP_service.GetTemplateModels()
print(template_models)
# All template models have the same handle. Need to use SetMCPCalc to have windPRO assign correct handles
ses.Models.TApiMCPModel = template_models
nan_to_skipvalue(mcp_calc)
calc_MCP_service.SetMCPCalc(mcp_calc)

# Checking the handles for the models.
mcp_calc = calc_MCP_service.GetMCPCalc()
for ses in mcp_calc.Sessions.TApiMcpSession[0].Models.TApiMCPModel:
    print('method: {}, handle:{}'.format(ses.Method, ses.Handle))

# Setting to 3 (the neural network model)
mcp_calc.Sessions.TApiMcpSession[0].Handle = 3

# Close and calculate
calculation_service.CloseCalc(save=True)
calculation_service.Calculate(handle=mcp_calc.Handle)

# Using MCP to make a new longer time series in the meteoobject for the measurement
# For this the calculationg needs to be open, since this calculation is started from within the GUI of the calculation
# in windPRO
calculation_service.OpenCalcForEdit(handle=resHandle)
mcp_calc = calc_MCP_service.GetMCPCalc()
calc_MCP_service.PredictToMeteo(sessionHandle=mcp_calc.Sessions.TApiMcpSession[0].Handle,
                                modelHandle=3,
                                inMeasMeteo=True,
                                newName='Scripting Long Term Corrected')

# Making wind statistics
calc_MCP_service.PredictToWindStat(sessionHandle=mcp_calc.Sessions.TApiMcpSession[0].Handle,
                                   modelHandle=3,
                                   wsFnName=os.path.join(working_dir, 'wind_stat_MCP.wws'))
calculation_service.CloseCalc(save=True)
